#pragma once

#include "z80ex-0.15/include/z80ex.h"
#include "devices/border/border.h"
#include "devices/mmanager/mmanager.h"
#include "devices/keyboard/keyboard.h"

#include "devices/trdos/trdos.h"
#include "devices/ay/ay.h"
#include "config.h"
#include "Attach_funcs.h"
#include "global_defs.h"
#include "snap_z80.h"

#define MAX_FRAME_TACTS 71680 
//#define MAX_FRAME_TACTS 10
#define INT_BEGIN (MAX_FRAME_TACTS-3606)

#define DRGB(r,g,b) (((r)<<0x10)|((g)<<8)|(b))
#define GETR(c) ((c)>>0x10)
#define GET_R GETR
#define GETG(c) ((c)>>8&0xFF)
#define GET_G GETG
#define GETB(c) ((c)&0xFF)
#define GET_B GETB 

extern SNDSAMPLE soundBuffer[SOUND_BUFFER_SIZE];
int soundBufferSize;
volatile int audioCnt; 

int colors[0x10] = {
	RGB(  0,   0,   0),
	RGB(  0,   0, 128),
	RGB(128,   0,   0),
	RGB(128,   0, 128),
	RGB(  0, 128,   0),
	RGB(  0, 128, 128),
	RGB(128, 128,   0),
	RGB(128, 128, 128),
	RGB(  0,   0,   0),
	RGB(  0,   0, 255),
	RGB(255,   0,   0),
	RGB(255,   0, 255),
	RGB(  0, 255,   0),
	RGB(  0, 255, 255),
	RGB(255, 255,   0),
	RGB(255, 255, 255)
};

unsigned short native_colors[0x10];
bool attributesHack = false; 

C_Border dev_border; 
C_Config config;
C_MemoryManager dev_mman; 
C_Keyboard dev_keyboard;
C_TrDos dev_trdos;
C_Ay dev_ay;

s_Params params; 
unsigned long long clk, currClk; 
extern bool drawKeyboard;
extern wchar_t filePath[128];

extern bool (* hnd_z80read[MAX_HANDLERS])(Z80EX_WORD, bool, Z80EX_BYTE&);
extern bool (* hnd_z80write[MAX_HANDLERS])(Z80EX_WORD, Z80EX_BYTE);
extern bool (* hnd_z80input[MAX_HANDLERS])(Z80EX_WORD, Z80EX_BYTE&);
extern bool (* hnd_z80output[MAX_HANDLERS])(Z80EX_WORD, Z80EX_BYTE);
extern void (* hnd_frameStart[MAX_HANDLERS])(void);
extern void (* hnd_afterFrameRender[MAX_HANDLERS])(void);
extern void (* hnd_reset[MAX_HANDLERS])(void); 
extern int cnt_z80read;
extern int cnt_z80write;
extern int cnt_z80input;
extern int cnt_z80output;
extern int cnt_frameStart;
extern int cnt_afterFrameRender;
extern int cnt_reset; 

void Action_Reset(void);
void Action_Exit(void);
void Action_ToggleKeyboard(void);
void Action_ResetTrDos(void);
void Action_MaxSpeed(void);
void Action_QuickLoad(void);
void Action_QuickSave(void);
void Action_AntiFlicker(void);
void Action_LoadFile(void);

C_Device* devs[] =
{
	&dev_border,
	&dev_keyboard,
	&dev_trdos,
	&dev_mman,
	&dev_ay,
	NULL
}; 

s_Action cfgActions[] =
{
	{"reset",		Action_Reset},
	{"exit",		Action_Exit},
	{"reset_trdos",		Action_ResetTrDos},
	{"max_speed",		Action_MaxSpeed},
	{"quick_load",		Action_QuickLoad},
	{"quick_save",		Action_QuickSave},
	{"anti_flicker",	Action_AntiFlicker},
	{"load_file",		Action_LoadFile},
	{"toggle_keyboard",		Action_ToggleKeyboard},
	{"",NULL}
}; 

class CSpeccy
{
protected:
	Z80EX_CONTEXT* cpu; 
	int drawFrame, m_skipFrameMax; 
	int frames;
	int skipFrame;
public:
	

	CSpeccy()
	{
		for (int i = 0; devs[i]; i++) devs[i]->Init(); 
		cpu = z80ex_create(&CSpeccy::ReadByte, NULL, &CSpeccy::WriteByte, NULL,	&CSpeccy::InputByte, NULL, &CSpeccy::OutputByte, NULL, &CSpeccy::ReadIntVec, NULL); 

		for (char i=0;i<0x10;i++)
			native_colors[i] = gd565::ColorrefToNative(colors[i]);

		z80ex_reset(cpu);
		config.GetInt(L"draw_every_nth_frame", &m_skipFrameMax);
		if (m_skipFrameMax==-1)
			m_skipFrameMax = 0;
		skipFrame = m_skipFrameMax;
		drawFrame = true;
	}

	void inc_frames()
	{
		frames++;
	}

	virtual void Render(GDBUFFERDESC* gdbd) {}

	int Step()
	{
		return z80ex_step(cpu);
	}

	bool LoadZ80Snap(wchar_t* filename)
	{
		return load_z80_snap(filename, cpu, dev_mman, dev_border);
	}

	void Reset()
	{
		int i;
		z80ex_reset(cpu);
		for (i = 0; i < cnt_reset; i++) hnd_reset[i]();
	}

	static Z80EX_BYTE ReadByte(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, int m1_state, void *userData)
	{
		int i;
		Z80EX_BYTE retval;

		for (i = 0; i < cnt_z80read; i++) {
			if (hnd_z80read[i](addr, m1_state, retval)) return retval;
		}

		return 0xFF;
	}

	static void WriteByte(Z80EX_CONTEXT *cpu, Z80EX_WORD addr, Z80EX_BYTE value, void *userData)
	{
		int i;
		for (i = 0; i < cnt_z80write; i++) {
			if (hnd_z80write[i](addr, value)) return;
		}
	}

	static Z80EX_BYTE InputByte(Z80EX_CONTEXT *cpu, Z80EX_WORD port, void *userData)
	{
		int i;
		Z80EX_BYTE retval;

		for (i = 0; i < cnt_z80input; i++) {
			if (hnd_z80input[i](port, retval)) return retval;
		}

		return 0xFF;
	}

	static void OutputByte(Z80EX_CONTEXT *cpu, Z80EX_WORD port, Z80EX_BYTE value, void *userData)
	{
		int i;
		for (i = 0; i < cnt_z80output; i++) {
			if (hnd_z80output[i](port, value)) return;
		}
	}

	static Z80EX_BYTE ReadIntVec(Z80EX_CONTEXT *cpu, void *userData)
	{
		return 0xFF;
	} 

};

extern CSpeccy* spec;

extern CMyApplication* pApplication;

void Action_Reset(void) 
{
	spec->Reset();
}

void Action_Exit(void)
{
	pApplication->Shutdown();
}

void Action_ResetTrDos(void)
{
	spec->Reset();
  dev_mman.OnOutputByte(0x7FFD, 0x10);
	dev_trdos.trdos = true;
}

void Action_ToggleKeyboard(void)
{
	drawKeyboard = !drawKeyboard;
}


void Action_MaxSpeed(void) {
//	params.maxSpeed = !params.maxSpeed;
}

void Action_QuickLoad(void) {
//	if (!load_z80_snap("snap.z80", cpu, dev_mman, dev_border)) StrikeMessage("Error loading snapshot");
}

void Action_QuickSave(void) {
//	save_z80_snap("snap.z80", cpu, dev_mman, dev_border);
}

void Action_AntiFlicker(void)
{
//	params.antiFlicker = !params.antiFlicker;
//	if (params.antiFlicker) SetMessage("AntiFlicker ON"); else SetMessage("AntiFlicker OFF");
}

void Action_LoadFile(void) {
//	FileDialog();
} 

class CSpeccy_QVGA_Crop : public CSpeccy
{
public:
  void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;

	/*		if (params.antiFlicker)
		{
			surf = scrSurf[sn];
			sn = 1 - sn;
		} else surf = screen;*/

	//		if (SDL_MUSTLOCK(surf)) {if (SDL_LockSurface(surf) < 0) return;}

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;
		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}
			short* buffer = (short*)gdbd->pBuffer/* + 480*72 + 80*/;
			if (drawFrame && !skipFrame)
			{
			//	drawFrame = false;
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=64 && line<256)	// in 320x240 screen (vertical)
					{
						scrLine = line - 64;
						clMod = rnClk % 224;

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256 || clMod<(36-5) || clMod>=(164-5-8))	// border
							{
						/*		cl = native_colors[dev_border.portFB & 7];
								pos = (clMod - (20-5)) * 2;
								scr = buffer + gdbd->dwWidth*scrLine + pos;  //PITCH
								*(scr++) = cl;
								*(scr) = cl;
								rnClk++;*/
								//scr++;
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
								scr = buffer + gdbd->dwWidth*scrLine + pos*8 ;  // +8

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos +1];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos +1];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								*(scr++) = (bt & 128 ? ci : cp);
								*(scr++) = (bt &  64 ? ci : cp);
								*(scr++) = (bt &  32 ? ci : cp);
								*(scr++) = (bt &  16 ? ci : cp);
								*(scr++) = (bt &   8 ? ci : cp);
								*(scr++) = (bt &   4 ? ci : cp);
								*(scr++) = (bt &   2 ? ci : cp);
								*(scr)   = (bt &   1 ? ci : cp);

								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}
	} 

};

class CSpeccy_QVGA_FullScreen : public CSpeccy
{
public:
  void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;
		bool isChetZnakomesto = true;

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;
		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}

			short* buffer = (short*)gdbd->pBuffer/* + 480*72 + 80*/;

			if (drawFrame && !skipFrame)
			{
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=40 && line<280)	// in 320x240 screen (vertical)
					{
						scrLine = line - 40;
						clMod = rnClk % 224;

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256 || clMod<(36-5) || clMod>=(164-5))	// border
							{
								cl = native_colors[dev_border.portFB & 7];
								pos = (clMod - (20-5)) * 2;
								//scr = buffer + gdbd->dwWidth*scrLine + pos;  //PITCH
								scr = buffer + (gdbd->dwHeight-1)*gdbd->dwWidth + scrLine - pos*gdbd->dwWidth;
								*(scr) = cl;
								scr -= gdbd->dwWidth;
								*(scr) = cl;
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
	//								scr = buffer + gdbd->dwWidth*scrLine + pos*8 - (pos >> 1) ;  // +8
								scr = buffer + (gdbd->dwHeight-1)*gdbd->dwWidth + scrLine - (32+8*pos)*gdbd->dwWidth;

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								*(scr) = (bt & 128 ? ci : cp); 		scr -= gdbd->dwWidth;
								*(scr) = (bt &  64 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr) = (bt &  32 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr) = (bt &  16 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr) = (bt &   8 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr) = (bt &   4 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr) = (bt &   2 ? ci : cp);		scr -= gdbd->dwWidth;
								*(scr)   = (bt &   1 ? ci : cp);
								isChetZnakomesto = !isChetZnakomesto;

								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}
	} 
};

class CSpeccy_VGA_Noresize : public CSpeccy
{
public:
	void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;

	/*		if (params.antiFlicker)
		{
			surf = scrSurf[sn];
			sn = 1 - sn;
		} else surf = screen;*/

	//		if (SDL_MUSTLOCK(surf)) {if (SDL_LockSurface(surf) < 0) return;}

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;

		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}

			short* buffer = (short*)gdbd->pBuffer + 480*72 + 80;
			if (drawFrame && !skipFrame)
			{
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=40 && line<280)	// in 320x240 screen (vertical)
					{
						scrLine = line - 40;
						clMod = rnClk % 224;

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256 || clMod<(36-5) || clMod>=(164-5))	// border
							{
								cl = native_colors[dev_border.portFB & 7];
								pos = (clMod - (20-5)) * 2;
								scr = buffer + gdbd->dwWidth*scrLine + pos;  //PITCH
								*(scr++) = cl;
								*(scr) = cl;
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
								scr = buffer + gdbd->dwWidth*scrLine + 32 + pos*8;

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos ];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos ];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								*(scr++) = (bt & 128 ? ci : cp);
								*(scr++) = (bt &  64 ? ci : cp);
								*(scr++) = (bt &  32 ? ci : cp);
								*(scr++) = (bt &  16 ? ci : cp);
								*(scr++) = (bt &   8 ? ci : cp);
								*(scr++) = (bt &   4 ? ci : cp);
								*(scr++) = (bt &   2 ? ci : cp);
								*(scr)   = (bt &   1 ? ci : cp);

								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}


	//		if (SDL_MUSTLOCK(surf)) SDL_UnlockSurface(surf);
	//		if (params.antiFlicker && drawFrame) AntiFlicker();
	} 

  
};

class CSpeccy_VGA_Resize : public CSpeccy
{
public:
	void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;

	/*		if (params.antiFlicker)
		{
			surf = scrSurf[sn];
			sn = 1 - sn;
		} else surf = screen;*/

	//		if (SDL_MUSTLOCK(surf)) {if (SDL_LockSurface(surf) < 0) return;}

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;

		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}

			short* buffer = (short*)gdbd->pBuffer;
			if (drawFrame && !skipFrame)
			{
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=64 && line<256)	// in 320x240 screen (vertical)
					{
						scrLine = line - 64;
						clMod = (rnClk % 224);

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256 || clMod<(36-5) || clMod>=(164-5))	// border
							{
								cl = native_colors[dev_border.portFB & 7];
								pos = (clMod - (20-5)) * 3;
								scr = buffer + 2*gdbd->dwWidth*scrLine + pos ;  //PITCH
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr) = cl;
								if (line>=64 && line <256)
								{
									scr += gdbd->dwWidth - 2;
									*(scr++) = cl;
									*(scr++) = cl;
									*(scr) = cl;
								}
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
								scr = buffer + 2*gdbd->dwWidth*scrLine + 32 + pos*13;

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos ];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos ];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								*(scr++) = (bt & 128 ? ci : cp);
								*(scr++) = (bt & 128 ? ci : cp);

								*(scr++) = (bt &  64 ? ci : cp);
								
								*(scr++) = (bt &  32 ? ci : cp);
								*(scr++) = (bt &  32 ? ci : cp);
								
								*(scr++) = (bt &  16 ? ci : cp);
								*(scr++) = (bt &  16 ? ci : cp);

								*(scr++) = (bt &   8 ? ci : cp);
								*(scr++) = (bt &   4 ? ci : cp);

								*(scr++) = (bt &   2 ? ci : cp);
								
								*(scr++) = (bt &   2 ? ci : cp);
								*(scr++) = (bt &   2 ? ci : cp);

								*(scr)   = (bt &   1 ? ci : cp);

							//	if (scrLine%4 != 0)
								{
									scr += gdbd->dwWidth - 12;
									*(scr++) = (bt & 128 ? ci : cp);
									*(scr++) = (bt & 128 ? ci : cp);

									*(scr++) = (bt &  64 ? ci : cp);
									
									*(scr++) = (bt &  32 ? ci : cp);
									*(scr++) = (bt &  32 ? ci : cp);
									
									*(scr++) = (bt &  16 ? ci : cp);
									*(scr++) = (bt &  16 ? ci : cp);

									*(scr++) = (bt &   8 ? ci : cp);
									*(scr++) = (bt &   4 ? ci : cp);

									*(scr++) = (bt &   2 ? ci : cp);
									
									*(scr++) = (bt &   2 ? ci : cp);
									*(scr++) = (bt &   2 ? ci : cp);

									*(scr)   = (bt &   1 ? ci : cp);
								}
								

								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}


	//		if (SDL_MUSTLOCK(surf)) SDL_UnlockSurface(surf);
	//		if (params.antiFlicker && drawFrame) AntiFlicker();
	} 	
};

class CSpeccy_QVGA_Resize : public CSpeccy
{
public:
	void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;
		bool isChetZnakomesto = true;

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;

		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}


			short* buffer = (short*)gdbd->pBuffer/* + 480*72 + 80*/;

			if (drawFrame && !skipFrame)
			{
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=64 && line<256)	// in 320x240 screen (vertical)
					{
						scrLine = line - 64;
						clMod = rnClk % 224;

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256 || clMod<(36-5) || clMod>=(164-5))	// border
							{
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
								scr = buffer + gdbd->dwWidth*scrLine + pos*8 - (pos >> 1) ;  // +8

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								if (isChetZnakomesto) *(scr++) = (bt & 128 ? ci : cp);
								*(scr++) = (bt &  64 ? ci : cp);
								*(scr++) = (bt &  32 ? ci : cp);
								*(scr++) = (bt &  16 ? ci : cp);
								*(scr++) = (bt &   8 ? ci : cp);
								*(scr++) = (bt &   4 ? ci : cp);
								*(scr++) = (bt &   2 ? ci : cp);
								*(scr)   = (bt &   1 ? ci : cp);
								isChetZnakomesto = !isChetZnakomesto;

								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}
	//		if (SDL_MUSTLOCK(surf)) SDL_UnlockSurface(surf);
	//		if (params.antiFlicker && drawFrame) AntiFlicker();
	} 
};

class CSpeccy_Smartphone_Resize : public CSpeccy
{
public:
	void Render(GDBUFFERDESC* gdbd)
	{
		unsigned long cmdClk, clMod, lastClk, rnClk;
		int line, scrLine, pos, cl;
		int zxLine, zxScreen, ci, cp, bt;
		int j, how;
		short *scr;
		bool intr;
		static int sn = 0;

	/*		if (params.antiFlicker)
		{
			surf = scrSurf[sn];
			sn = 1 - sn;
		} else surf = screen;*/

	//		if (SDL_MUSTLOCK(surf)) {if (SDL_LockSurface(surf) < 0) return;}

		currClk = 0;
		lastClk = 0;
		intr = true;

		skipFrame--;
		if (skipFrame==-1)
			skipFrame = m_skipFrameMax;
		drawFrame = true;

		while (currClk < MAX_FRAME_TACTS)
		{
			cmdClk = z80ex_step(cpu);
			currClk += cmdClk;
			clk += cmdClk;

			if (params.maxSpeed && z80ex_doing_halt(cpu))
			{
				cmdClk = (MAX_FRAME_TACTS - currClk - 1) / 4 + 1;
				currClk += cmdClk * 4;
				clk += cmdClk * 4;
				break;
			}

			short* buffer = (short*)gdbd->pBuffer/* + 480*72 + 80*/;

			if (drawFrame && !skipFrame)
			{
				for (rnClk = lastClk; rnClk < currClk;)
				{
					line = rnClk / 224;

					if (line>=50 && line<270)	// in 320x240 screen (vertical)
					{
						scrLine = line - 50;
						clMod = rnClk % 224;

						if (clMod>=(20-5) && clMod<(180-5))	// in 320x240 screen (horizontal)
						{
							if (line<64 || line>=256)	// border
							{
								cl = native_colors[dev_border.portFB & 7];
								pos = (((clMod - (20-5))*2));
								if (pos < 176) // Width
								{
									scr = buffer + gdbd->dwWidth*scrLine + pos;  //PITCH
									*(scr++) = cl;
									*(scr) = cl;
								}
								rnClk++;
							}
							else if (clMod<(36-5) || clMod>=(164-5))
							{
								cl = native_colors[dev_border.portFB & 7];
								if (clMod <(36-5))
									scr = buffer + gdbd->dwWidth*scrLine;
								else
									scr = buffer + gdbd->dwWidth*scrLine + 168;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr++) = cl;
								*(scr) = cl;
								rnClk++;
							}
							else
							{
								pos = (clMod - (36-5)) / 4;
								zxLine = line - 64;
								scr = buffer + gdbd->dwWidth*scrLine + pos*5 + 8;

								zxScreen = (dev_mman.port7FFD & 8) ? RAM_BANK7 : RAM_BANK5;
								bt = dev_mman.ram[ zxScreen + ((zxLine & 0xC0) << 5) + ((zxLine & 7) << 8) + ((zxLine & 0x38) << 2) + pos];
								cl = dev_mman.ram[ zxScreen + 0x1800 + ((zxLine & 0xF8) << 2) + pos];

								if ((frames&32) && (cl&128))
								{
									cp = native_colors[((cl & 64) >> 3) | (cl & 7)];
									ci = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}
								else
								{
									ci = native_colors[((cl & 64) >> 3) | (cl & 7)];
									cp = native_colors[((cl & 64) >> 3) | ((cl >> 3) & 7)];
								}

								if (attributesHack)
								{
									ci = DRGB(0,0,0);
									cp = DRGB(255,255,255);
								}

								//if (isChetZnakomesto)
								*(scr++) = (((bt &  128) | (bt &  64)) ? ci : cp); 
								*(scr++) = (bt &  32 ? ci : cp);
								*(scr++) = (((bt &   16) | (bt & 8 ))? ci : cp);
								*(scr++) = (bt &   4 ? ci : cp);
								*(scr) = (((bt &   2) | (bt & 1)) ? ci : cp); 
								
								if (clMod <= 160) rnClk += 4;
								else rnClk++;
							}
						} else rnClk++;
					} else rnClk++;
				}
				lastClk = rnClk;
			}

			if (currClk>=INT_BEGIN && currClk < (INT_BEGIN  + 32/*INT len*/))
			{
				cmdClk = z80ex_int(cpu);
				currClk += cmdClk;
				clk += cmdClk;
			}
		}
	//		if (SDL_MUSTLOCK(surf)) SDL_UnlockSurface(surf);
	//		if (params.antiFlicker && drawFrame) AntiFlicker();
	} 
};