#include "trdos.h"
#include "../../mapfile.h"
#include "../../exceptions.h"
#include "../../lib_wd1793/unreal_wd1793.h"
#include "../mmanager/mmanager.h"
#include "../../config.h"
#include "../../Attach_funcs.h"
#include <comutil.h>

extern C_MemoryManager dev_mman;
extern C_Config config;
extern unsigned long long clk, currClk; 
extern _bstr_t filePathString;

bool C_TrDos::trdos;
BYTE C_TrDos::rom[0x4000];
bool C_TrDos::wd_flag;

void C_TrDos::ReadFile(void)
{
	int i;
	C_MapFile fl;
	wchar_t *fname;

	if (!config.GetString(L"romTrDos", &fname)) throw C_E(E_General, "root/Roms/TrDos not found in config");
	fl.Read(fname);
	fl.ReadBlock(rom, 0x4000);
	fl.Close();
	wchar_t* buf = new wchar_t[128];
	if (config.GetString(L"diska", &buf))
		wd1793_load_dimage(filePathString+buf, 0);
	if (config.GetString(L"diskb", &buf))
		wd1793_load_dimage(filePathString+buf, 1);
	if (config.GetString(L"diskc", &buf))
		wd1793_load_dimage(filePathString+buf, 2);
	if (config.GetString(L"diskd", &buf))
		wd1793_load_dimage(filePathString+buf, 3);		
	delete[] buf;
}

void C_TrDos::Init(void)
{
	ReadFile();
	AttachZ80ReadHandler(OnReadByte);
	AttachZ80InputHandler(OnInputByte);
	AttachZ80OutputHandler(OnOutputByte);
	AttachResetHandler(OnReset);
	trdos = false;
}

void C_TrDos::Close(void)
{
}

bool C_TrDos::OnReadByte(Z80EX_WORD addr, bool m1, Z80EX_BYTE &retval)
{
	if (m1)
	{
		if ((addr & 0xFF00)==0x3D00 && (dev_mman.port7FFD & 0x10)) trdos = true;
		else
		if (addr > 0x3FFF) trdos = false;
	}

	if (trdos && addr<0x4000)
	{
		retval = rom[addr];
		return true;
	}

	return false;
}

bool C_TrDos::OnInputByte(Z80EX_WORD port, Z80EX_BYTE &retval)
{
	int err;
	int lport = port & 0xFF;
	if (trdos && (lport==0xFF || lport==0x7F || lport==0x1F || lport==0x3F || lport==0x5F))
	{
		wd_flag = true;
		retval = wd1793_in(lport, clk, &err);
		return true;
	}
	return false;
}

bool C_TrDos::OnOutputByte(Z80EX_WORD port, Z80EX_BYTE value)
{
	int err;
	int lport = port & 0xFF;
	if (trdos && (lport==0xFF || lport==0x7F || lport==0x1F || lport==0x3F || lport==0x5F))
	{
		wd_flag = true;
		wd1793_out(lport, value, clk, &err);
		return true;
	}
	return false;
}

void C_TrDos::OnReset(void)
{
	trdos = false;
}
