#ifndef STRICT
#define STRICT
#endif

#pragma comment(lib,"comsuppw.lib")
#pragma comment(lib,"GapiDraw.lib")
#ifdef _WIN32_WCE
#pragma comment(lib,"../sdl_WinCE.lib")
#else
#pragma comment(lib,"../sdl.lib")
#pragma comment(lib,"../sdlmain.lib")
#endif

#include "MyApplication.h"
#include "GapiDrawExtension.h"
#include "resource.h"
#include "bin.h"
#include "lib_sdl\sdl.h"
#include "devices/keyboard/keyboard.h"
#include "defines.h"
#include "speccy.h"
#include "lib_ay\sndcounter.h"
#include "lib_wd1793\unreal_wd1793.h"
#include "ScreenKeyboard.h"
#include "wavplay/wavplay.h"

int GapiDrawMain(HINSTANCE hInst);

CSpeccy* spec;
CMyApplication* pApplication;
extern C_Keyboard dev_keyboard;

bool disableSound; 
SDL_AudioSpec *hardware_spec; 
unsigned int btick; 
extern unsigned frameSamples;
extern SNDCOUNTER sndCounter; 
extern unsigned soundPtr;

bool drawKeyboard = true;

#define FRAME_WAIT_MS 20
#define MAX_SPEED_FRAME_SKIP 64 

// All files should must be placed in the application directory, or you should enter full path



//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Application Entry point
//-----------------------------------------------------------------------------
#ifdef _WIN32_WCE
int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE /*hPrevInst*/, LPWSTR /*pCmdLine*/, int /*nCmdShow*/)
{
	return GapiDrawMain(hInst);
}
#else // _WIN32_WCE
int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE /*hPrevInst*/, LPSTR /*pCmdLine*/, int /*nCmdShow*/)
{
	return GapiDrawMain(hInst);
}
#endif // _WIN32_WCE

void TruncateProgramPath()
{
	int length = (int)wcslen(filePath);
	for (int z = length-1; z>0; z--)
		if (filePath[z]=='\\')
		{
			filePath[++z] = '\0';
			break;
		}
	filePathString = filePath;
}

void AudioCallback(void *userdata, Uint8 *stream, int len)
{
	if (disableSound || params.maxSpeed)
	{
		while (len--) *(stream++) = 0;
	}
	else
	{
		for (len /= 4; len--;)
		{
		
			*((WORD *)stream) = soundBuffer[audioCnt].ch.left;
			stream += 2;
			*((WORD *)stream) = soundBuffer[audioCnt].ch.right;
			stream += 2;
			
			soundBuffer[audioCnt].ch.left=0;
			soundBuffer[audioCnt].ch.right=0;

			audioCnt++;
			if(audioCnt >= SOUND_BUFFER_SIZE) audioCnt = 0;
		}
	}
} 

void InitAudio(void)
{
	SDL_AudioSpec *desired, *obtained;

	//[boo]
	memset(soundBuffer,0,SOUND_BUFFER_SIZE);	
	//[boo] - samples per frame (rounded)
	frameSamples = ((unsigned long)71680/*frame*/*44100/*sound freq*/) / 3500000 /*Z80 clock*/;
	
	if (params.sdl_sound)
	{
		desired = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));
		obtained = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));

		desired->freq = 44100;
		desired->format = AUDIO_S16LSB;
		desired->channels = 2;
		desired->samples = AUDIO_HW_BUFFER;
		desired->callback = AudioCallback;

		if (SDL_OpenAudio(desired, obtained) < 0) 
			StrikeError("Couldn't open audio: %s\n", SDL_GetError());

		free(desired);
		hardware_spec = obtained;

		audioCnt = frameSamples; //[boo]
		
		SDL_PauseAudio(0);
	}
	else
	{
		wav_start(frameSamples *2*2 /*[boo] sound frame size in bytes*/);
	}
}

void FreeAudio(void)
{
	if (params.sdl_sound)
	{
		SDL_CloseAudio();
		free(hardware_spec);
	}
	else
	{
//		wav_stop();
	}
} 

//-----------------------------------------------------------------------------
// Name: GapiDrawMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Exit the loop by calling Shutdown().
//-----------------------------------------------------------------------------
int GapiDrawMain(HINSTANCE hInst)
{
	HMODULE hMod = GetModuleHandle(0);
	GetModuleFileName(hMod, filePath, 128);
	TruncateProgramPath();

	config.Init(L"config.ini");

	GDAPPCONFIG configGDA;
	::ZeroMemory(&configGDA, sizeof(GDAPPCONFIG));

	configGDA.hInstance           = hInst;
	configGDA.pAppTitle           = _T("Minimal application");
	configGDA.dwWindowIcon        = IDI_APP;
	configGDA.dwDisplayFlags      = 0;
	configGDA.dwDisplayMode       = GDDISPMODE_NORMAL; // No display rotation
	configGDA.dwMaxFPS            = 50; // Limit number of calls to ProcessNextFrame each second to 30

	params.isVga = false;
	params.isConfigVga = config.GetBool(L"vga");

	// The following flags are only relevant if compiled with Embedded Visual C++ 4.0 or later
#if defined(_WIN32_WCE) && (_WIN32_WCE >= 420)
	if (params.isConfigVga)
		configGDA.dwDisplayFlags     |= GDDISPLAY_HI_RES_AWARE; // Enable VGA resolution
	int xScr = GetSystemMetrics(SM_CXSCREEN), yScr = GetSystemMetrics(SM_CYSCREEN);
	if (xScr==640 || yScr==640) 
		params.isVga = true;
#else
	params.isVga = true;
#endif

	// The following flags are only relevant on stationary PCs
//#ifdef _DEBUG
	configGDA.dwDisplayFlags     |= GDDISPLAY_WINDOW;
	configGDA.dwWindowBorder      = WS_DLGFRAME; // Set bordertype to WS_DLGFRAME to prevent window resizing
	configGDA.dwDisplayWidth      = 480;
	configGDA.dwDisplayHeight     = 640;
	configGDA.dwDisplayZoomWidth  = 480;
	configGDA.dwDisplayZoomHeight = 640;
/*#else
	config.dwDisplayFlags     |= GDDISPLAY_FULLSCREEN;
	config.dwDisplayWidth      = 640;
	config.dwDisplayHeight     = 480;
	config.dwDisplayHz         = 75;
#endif
*/
	// Create the CGapiApplication subclass
	CMyApplication* pMyApp = new CMyApplication(configGDA);
	
	params.sdl_sound = false;
	params.sound = false;
	params.timing_snd = false;

	InitAudio(); 

//	spec = new CSpeccy_Smartphone_Resize();

	wchar_t* buf;
	config.GetString(L"video_mode", &buf);
	if (buf==0)
		spec = new CSpeccy_QVGA_Crop();
	else
	if (wcscmp(buf, L"qvga_crop")==0) 
		spec = new CSpeccy_QVGA_Crop();
	else if (wcscmp(buf, L"qvga_resize")==0)
		spec = new CSpeccy_QVGA_Resize();
	else if (wcscmp(buf, L"qvga_fullscreen")==0)
		spec = new CSpeccy_QVGA_FullScreen();
	else if (wcscmp(buf, L"vga_resize")==0)
		spec = new CSpeccy_VGA_Resize();
	else if (wcscmp(buf, L"vga_noresize")==0)
		spec = new CSpeccy_VGA_Noresize();
	else if (wcscmp(buf, L"smartphone")==0)
		spec = new CSpeccy_Smartphone_Resize();
	else
		spec = new CSpeccy_QVGA_Crop();

	LPWSTR commandLine = GetCommandLine();
	#ifndef _WIN32_WCE
	while (*commandLine != L' ' && *commandLine != 0) commandLine++;
	commandLine++;
	if (*commandLine==L'"') 
	{
		commandLine++;
		LPWSTR cl = commandLine;
		while(*cl!=L'"' && *cl!=0) cl++;
		*cl = 0;
	}
	#endif
	if (*commandLine != 0) wd1793_load_dimage((LPWSTR)commandLine, 0);

	btick = SDL_GetTicks() + FRAME_WAIT_MS; 

	// Start the main loop
	HRESULT hr = -1;
	if (pMyApp)
	{
		pApplication = pMyApp;
		hr = pMyApp->Run();
		delete pMyApp;
	}

	// Exit application
	return hr;
}

CMyApplication::CMyApplication(const GDAPPCONFIG& config) : CGapiApplication(config)
{
}

CMyApplication::~CMyApplication()
{
}

HRESULT CMyApplication::InitInstance()
{
	m_screenKeyboard = 0;
	return S_OK;
}

HRESULT CMyApplication::ExitInstance()
{
	return S_OK;
}

HRESULT CMyApplication::OnMinimize()
{
	return S_OK;
}

HRESULT CMyApplication::OnRestore()
{
	return S_OK;
}

HRESULT CMyApplication::CreateSysMemSurfaces(CGapiDisplay* pDisplay, HINSTANCE hInstance)
{
	return S_OK;
}

HRESULT CMyApplication::CreateVidMemSurfaces(CGapiDisplay* pDisplay, HINSTANCE hInstance)
{
	return S_OK;
}

HRESULT CMyApplication::ProcessNextFrame(CGapiSurface* pBackBuffer, DWORD dwFlags)
{
	// Clear the back buffer
	//pBackBuffer->FillRect(NULL, RGB(128, 128, 128), 0, NULL);

	if (!m_screenKeyboard)
	{
		m_screenKeyboard = new CScreenKeyboard();
		m_screenKeyboard->Init(GetGlobal(), params.isConfigVga);		
	}

	GDBUFFERDESC gdbd;
	pBackBuffer->GetBuffer(&gdbd);

	if (params.sdl_sound && params.timing_snd && params.sound && !params.maxSpeed && soundBufferSize>AUDIO_BUFFER_SIZE)
	{
		int ayCntX;
		do
		{
			SDL_Delay(1);
			ayCntX = dev_ay.ayCnt;
			if (ayCntX < audioCnt) ayCntX += soundBufferSize;
		} while ((ayCntX-audioCnt) > AUDIO_BUFFER_SIZE);
	} 

	dev_trdos.wd_flag = false; 
/*	if (params.isConfigVga)
		spec->Render_Pentagon_VGA_Resize(&gdbd);
	else
		spec->Render_Pentagon_Smartphone_Resize(&gdbd);*/
	spec->Render(&gdbd);
	spec->inc_frames();
	pBackBuffer->ReleaseBuffer();

/*	sndCounter.begin();
//	sndCounter.count(C_Ay::ayChip);
	sndCounter.count(C_Border::spkRender);
	sndCounter.end(soundPtr,frameSamples);
	
*/
	if (drawKeyboard)
		m_screenKeyboard->Draw(pBackBuffer);
	
	for (int i = 0; i < cnt_afterFrameRender; i++) hnd_afterFrameRender[i](); 

	if(!params.sdl_sound)
	{
			sndCounter.begin();
		
			sndCounter.count(C_Ay::ayChip);
			sndCounter.count(C_Border::spkRender);
		
			sndCounter.end(soundPtr,frameSamples);
		
	//		wav_play(soundBuffer, soundPtr, frameSamples, SOUND_BUFFER_SIZE);
	}

	return S_OK;
}

HRESULT CMyApplication::KeyDown(DWORD dwKey, GDKEYLIST& keylist)
{/*
 
 today - 91
back - 27
 - 112
 - 113
 - 114
 - 115
121 - 
118 - volume down
117 - --//-- up
 - 

 */
	dev_keyboard.OnKeyDown(dwKey);
	return S_OK;
}

HRESULT CMyApplication::Shutdown()
{
	return CGapiApplication::Shutdown();
}


HRESULT CMyApplication::KeyUp(DWORD dwKey, GDKEYLIST& keylist)
{
	dev_keyboard.OnKeyUp(dwKey);
	return S_OK;
}

HRESULT CMyApplication::StylusDown(POINT p)
{
	m_screenKeyboard->ProcessStylusDown(p);
	return S_OK;
}

HRESULT CMyApplication::StylusUp(POINT p)
{
	m_screenKeyboard->ProcessStylusUp(p);
	return S_OK;
}

HRESULT CMyApplication::StylusDblClk(POINT p)
{
	return S_OK;
}

HRESULT CMyApplication::StylusMove(POINT p)
{
	return S_OK;
}